/*
 * Copyright (c) 2009 Matthias Bach,
 *                    matthias.bach@kip.uni-heidelberg.de
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <IL/il.h>

#include <complex>
#include <iostream>
#include <cmath>

using namespace std;

/**
 * Check whether iteration from the give start value will converge to 0.
 */
unsigned char mandelbrot( float real, float imag,
	 const float max_abs, const unsigned char max_iter )
{
	complex<float> a( real, imag );
	complex<float> z = a;

	unsigned int iter = 0;
	while( abs( z ) < max_abs && iter < max_iter )
	{
		z = z * z + a;
		++iter;
	}

	return iter;
}

int main(int argc, char** argv)
{
	//
	// PREPARATION
	//
	ilInit();

	// Problem dimension
	int cols = 786, rows = 786;
	// Origin
	float x0 = -0.5f;
	float y0 = 0.0f;
	// Stepsize
	float sx = 2.0f / cols;
	float sy = 2.0f / rows;

	// Place for the result
	unsigned char data[ rows ][ cols ][ 3 ];

	// Mandelbrot calculation

	for( int i = 0; i < cols; ++i )
	{
		float x = x0 + sx * ( (float)i - cols / 2.0f );
		for( int j = 0; j < rows; ++j )
		{
			float y = y0 + sy * ( (float)j - rows / 2.0f);

			unsigned char val = 255 - mandelbrot( x, y, 2.0f, 255 );

			// some nice colors
			data[ j ][ i ][ 0 ] = sinf( val * 2.0f ) * 256.0f;
			data[ j ][ i ][ 1 ] = cosf( val * 2.0f ) * 256.0f;
			data[ j ][ i ][ 2 ] = val;
		}
	}


	// End of Mandelbrot calculation
	
	//
	// EXPORT RESULT AS PNG
	// 
	ILuint imageName;
	ilGenImages( 1, &imageName );
	ilBindImage( imageName );

	ilTexImage( cols, rows, 1, 3, IL_RGB, IL_UNSIGNED_BYTE, data );

	ilEnable( IL_FILE_OVERWRITE );
	ilSave( IL_PNG, "mandelbrot.png" );

	ilDeleteImages( 1, &imageName );
}

