/*
 * Copyright (c) 2010 Matthias Bach,
 *                    bach@compeng.uni-frankfurt.de
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <IL/il.h>
#include <assert.h>
#include <iostream>
#include <CL/cl.h>

#define str(s) #s
#define strfy(foo) str(foo)

#define SOURCE_FILE_NAME "mandelbrot.cl"
#define OPENCL_COMPILER_OPTIONS "-cl-mad-enable"
#define quit(msg) {printf(msg "\n");return(1);}

using namespace std;

int main(int argc, char** argv)
{
	//
	// PREPARATION
	//
	ilInit();

	// Problem dimension
	cl_uint cols = 768, rows = 768;

	// Place for the result
	cl_uchar3 data[ rows ][ cols ];

	cl_int ocl_error;

	//Query platform count
	cl_uint num_platforms;
	if (clGetPlatformIDs(0, NULL, &num_platforms) != CL_SUCCESS) quit("Error getting OpenCL Platform Count");
	if (num_platforms == 0) quit("No OpenCL Platform found");
	printf("%d OpenCL Platforms found\n", num_platforms);
	
	//Query platforms
	cl_platform_id* platforms = new cl_platform_id[num_platforms];
	if (platforms == NULL) quit("Memory allocation error");
	if (clGetPlatformIDs(num_platforms, platforms, NULL) != CL_SUCCESS) quit("Error getting OpenCL Platforms");

	//Query and display platform information
	for (int i = 0;i < num_platforms;i++)
	{
		char platform_profile[64], platform_version[64], platform_name[64], platform_vendor[64];
		clGetPlatformInfo(platforms[i], CL_PLATFORM_PROFILE, 64, platform_profile, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 64, platform_version, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 64, platform_name, NULL);
		clGetPlatformInfo(platforms[i], CL_PLATFORM_VENDOR, 64, platform_vendor, NULL);
		printf("Platform %d: (%s %s) %s %s\n", i, platform_profile, platform_version, platform_vendor, platform_name);
	}

	//Set first platform as default platform
	cl_platform_id platform = platforms[0];
	delete[] platforms;

	//Query device count for this platform
	cl_uint num_devices;
	clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 0, NULL, &num_devices);
	if (num_devices == 0) quit("No OpenCL device for this platform found");
	printf("%d OpenCL devices found for this platform\n", num_devices);

	//Query devices
	cl_device_id* devices = new cl_device_id[num_devices];
	if (devices == NULL) quit("Memory allocation error");
	if (clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, num_devices, devices, NULL) != CL_SUCCESS) quit("Error getting OpenCL devices");

	//Query and display device information
	for (int i = 0;i < num_devices;i++)
	{
		char device_vendor[64], device_name[64];
		clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 64, device_name, NULL);
		clGetDeviceInfo(devices[i], CL_DEVICE_VENDOR, 64, device_vendor, NULL);
		printf("Device %d: %s %s\n", i, device_vendor, device_name);
	}

	//Set first device as default device
	cl_device_id device = devices[0];
	delete[] devices;

	//Create OpenCL context
	cl_context context = clCreateContext(NULL, 1, &device, NULL, NULL, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating OpenCL context");

	//Create OpenCL command_queue
	cl_command_queue command_queue = clCreateCommandQueue(context, device, 0, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating OpenCL command queue");

	//Read Sourcecode from file
	unsigned int filesize;
	FILE* pFile;
	//Open file
	if ((pFile = fopen(SOURCE_FILE_NAME, "rb")) == NULL) quit("Error opening OpenCL source code file");
	//Optain File Size
	fseek (pFile , 0 , SEEK_END);
	filesize = ftell (pFile);
	rewind (pFile);
	//Read file
	char* sourceCode = new char[filesize + 1];
	if (fread(sourceCode, 1, filesize, pFile) != filesize) quit("Error reading source file");
	//Make sourceCode 0-terminated
	sourceCode[filesize] = 0;
	fclose(pFile);

	//Create OpenCL program object
	cl_program program = clCreateProgramWithSource(context, 1, (const char**) &sourceCode, NULL, &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating program object");
	delete[] sourceCode;

	//Compile program
	ocl_error = clBuildProgram(program, 1, &device, OPENCL_COMPILER_OPTIONS, NULL, NULL);
	if (ocl_error != CL_SUCCESS) quit("Error compiling program");

	//Create kernel
	cl_kernel kernel = clCreateKernel(program, "mandelbrot_gpu", &ocl_error);
	if (ocl_error != CL_SUCCESS) quit("Error creating kernel");

	// Allocate some memory on the device
	cl_mem d_data;
	d_data = clCreateBuffer( context, CL_MEM_WRITE_ONLY, sizeof( data ), NULL, &ocl_error ); 
	if (ocl_error != CL_SUCCESS) quit("Error allocating device memory");
	
	// set the kernel arguments
	ocl_error = clSetKernelArg( kernel, 0, sizeof( cl_mem ), &d_data );
	if (ocl_error != CL_SUCCESS) quit("Error setting kernel argument 0");
	ocl_error = clSetKernelArg( kernel, 1, sizeof( cl_uint ), &rows ); 
	if (ocl_error != CL_SUCCESS) quit("Error setting kernel argument 1");
	ocl_error = clSetKernelArg( kernel, 2, sizeof( cl_uint ), &cols ); 
	if (ocl_error != CL_SUCCESS) quit("Error setting kernel argument 2");

	// Make the device execute our code
	size_t workitems[2] = {cols,rows};
	size_t local_workitems[2] = {16,16};
	ocl_error = clEnqueueNDRangeKernel( command_queue, kernel, 2, NULL, workitems, local_workitems, 0, NULL, NULL ); 
	if (ocl_error != CL_SUCCESS) quit("Error executing kernel");

	// and queue reading
	ocl_error = clEnqueueReadBuffer( command_queue, d_data, false, 0, sizeof(data), &data, 0, NULL, NULL ); 
	if (ocl_error != CL_SUCCESS) quit("Error reading data from device");

	// cleanup
	clReleaseKernel( kernel );
	clReleaseMemObject( d_data );
	clReleaseCommandQueue( command_queue );

	unsigned char out_data[ rows ][ cols ][ 3 ];

	for( int j = 0; j < rows; ++j )
	{
		for( int i = 0; i < cols; ++i )
		{
			out_data[j][i][0] = data[j][i].x;
			out_data[j][i][1] = data[j][i].y;
			out_data[j][i][2] = data[j][i].z;
		}
	}

	//
	// EXPORT RESULT AS PNG
	// 
	ILuint imageName;
	ilGenImages( 1, &imageName );
	ilBindImage( imageName );

	ilTexImage( cols, rows, 1, 3, IL_RGB, IL_UNSIGNED_BYTE, out_data );

	ilEnable( IL_FILE_OVERWRITE );
	ilSave( IL_PNG, "mandelbrot.png" );

	ilDeleteImages( 1, &imageName );
}
